import jax
import numpy as np
import pytest

from jaxoplanet.experimental.starry import Map, Ylm
from jaxoplanet.experimental.starry.light_curves import light_curve, map_light_curve
from jaxoplanet.experimental.starry.orbit import SurfaceMapSystem
from jaxoplanet.orbits import keplerian
from jaxoplanet.test_utils import assert_allclose


@pytest.mark.parametrize("deg", [2, 5, 10])
@pytest.mark.parametrize("u", [[], [0.1], [0.2, 0.1]])
def test_compare_starry(deg, u):
    starry = pytest.importorskip("starry")
    starry.config.lazy = False

    # map
    inc = np.pi / 2
    np.random.seed(deg)
    y = Ylm.from_dense(np.random.randn((deg + 1) ** 2))
    map = Map(y=y, u=u, inc=inc)

    # occultor
    yo = np.linspace(-3, 3, 1000)
    ro = 0.01
    zo = 2.0
    xo = 0.0

    # starry
    ms = starry.Map(ydeg=deg, udeg=len(u), inc=np.rad2deg(inc))
    ms[:, :] = y.todense()
    if len(u) > 0:
        ms[1:] = u
    expected = ms.flux(xo=xo, yo=yo, ro=ro, zo=zo)

    # jaxoplanet
    calc = jax.vmap(map_light_curve, in_axes=(None, None, None, 0, None, None))(
        map, ro, xo, yo, zo, 0.0
    )

    assert_allclose(calc, expected)


@pytest.fixture(
    params=[
        {
            "central": keplerian.Central(
                mass=1.3,
                radius=1.1,
            ),
            "surface_map": Map(
                y=Ylm.from_dense(
                    [1, 0.005, 0.05, 0.09, 0.0, 0.1, 0.03, 0.04, 0.4, 0.2, 0.1]
                ),
                inc=0.9,
                obl=0.3,
                period=1.2,
                u=(0.1, 0.1),
            ),
            "body": {
                "radius": 0.5,
                "mass": 0.1,
                "period": 1.5,
                "surface_map": Map(
                    y=Ylm.from_dense([1, 0.005, 0.05, 0.09, 0.0, 0.1, 0.03]),
                    inc=-0.3,
                    period=0.8,
                    u=(0.2, 0.3),
                ),
            },
        },
        {
            "central": keplerian.Central(
                mass=1.3,
                radius=1.1,
            ),
            "surface_map": Map(
                y=Ylm.from_dense(np.hstack([1, 0.005, 0.05, 0.09, 0.0, 0.1, 0.03])),
                period=1.2,
                u=(0.1, 0.1),
            ),
            "body": {
                "radius": 0.5,
                "mass": 0.1,
                "period": 1.5,
            },
        },
        {
            "central": keplerian.Central(
                mass=0.3,
                radius=0.1,
            ),
            "body": {
                "radius": 1.5,
                "mass": 1.1,
                "period": 1.5,
                "surface_map": Map(
                    y=Ylm.from_dense([1, 0.005, 0.05, 0.09, 0.0, 0.1, 0.03]),
                    period=1.2,
                    u=(0.1, 0.1),
                ),
            },
        },
    ]
)
def keplerian_system(request):
    return SurfaceMapSystem(
        request.param["central"], request.param.get("surface_map", None)
    ).add_body(**request.param["body"])


def test_compare_starry_system(keplerian_system):
    def extract_starry_map_kwargs(jaxoplanet_map):
        if jaxoplanet_map is None:
            # starry's simplest case: a limb darkened map
            map_kwargs = dict(ydeg=0, rv=False, reflected=False)
        else:
            map_kwargs = dict(
                ydeg=jaxoplanet_map.ydeg,
                udeg=jaxoplanet_map.udeg,
                inc=np.rad2deg(jaxoplanet_map.inc),
                obl=np.rad2deg(jaxoplanet_map.obl),
                amp=jaxoplanet_map.amplitude,
            )
        return map_kwargs

    starry = pytest.importorskip("starry")
    starry.config.lazy = False

    # jaxoplanet system
    central = keplerian_system.central
    central_map = keplerian_system.surface_map
    body = keplerian_system.bodies[0]
    body_map = keplerian_system.surface_maps[0]

    time = np.linspace(-1.5, 1.0, 300)
    jaxoplanet_flux = light_curve(keplerian_system)(time)

    # starry system
    central_map_kwargs = extract_starry_map_kwargs(central_map)
    pri = starry.Primary(
        starry.Map(**central_map_kwargs),
        r=central.radius.magnitude,
        m=central.mass.magnitude,
        prot=0 if central_map is None else central_map.period,
    )
    if central_map and central_map.u:
        pri.map[1:] = central_map.u
    if central_map and central_map.deg > 0:
        pri.map[1:, :] = np.asarray(central_map.y.todense())[1:]

    body_map_kwargs = extract_starry_map_kwargs(body_map)
    sec = starry.Secondary(
        starry.Map(**body_map_kwargs),
        r=body.radius.magnitude,
        m=body.mass.magnitude,
        porb=body.period.magnitude,
        prot=0 if body_map is None else body_map.period,
    )
    if body_map and body_map.u:
        sec.map[1:] = body_map.u
    if body_map and body_map.deg > 0:
        sec.map[1:, :] = np.asarray(body_map.y.todense())[1:]

    starry_system = starry.System(pri, sec)
    starry_flux = starry_system.flux(time, total=False)

    # compare summed jaxponet fluxes with the corresponding starry fluxes
    if body_map is None:
        assert_allclose(jax.numpy.squeeze(jaxoplanet_flux), starry_flux[0])
    elif central_map is None:
        assert_allclose(jax.numpy.squeeze(jaxoplanet_flux), starry_flux[1])
    else:
        assert_allclose(jax.numpy.squeeze(jaxoplanet_flux), np.sum(starry_flux, axis=0))

    # compare individual jaxpolanet fluxes with the corresponding starry fluxes
    # for jaxoplanet_flux_item, starry_flux_item in zip(jaxoplanet_flux, starry_flux):
    #     if jaxoplanet_flux_item is not None:
    #         assert_allclose(jax.numpy.squeeze(jaxoplanet_flux_item), starry_flux_item)



from collections.abc import Callable
from functools import partial
from typing import Optional

import jax
import jax.numpy as jnp
import numpy as np
import scipy
from jax.experimental import sparse

from jaxoplanet.experimental.starry.basis import A1, U0, A2_inv
from jaxoplanet.experimental.starry.maps import Map
from jaxoplanet.experimental.starry.orbit import SurfaceMapSystem
from jaxoplanet.experimental.starry.pijk import Pijk
from jaxoplanet.experimental.starry.rotation import left_project
from jaxoplanet.experimental.starry.solution import solution_vector
from jaxoplanet.light_curves.utils import vectorize
from jaxoplanet.types import Array, Quantity
from jaxoplanet.units import quantity_input, unit_registry as ureg


def light_curve(
    system: SurfaceMapSystem,
) -> Callable[[Quantity], tuple[Optional[Array], Optional[Array]]]:
    central_bodies_lc = jax.vmap(map_light_curve, in_axes=(None, 0, 0, 0, 0, None))

    @partial(system.surface_map_vmap, in_axes=(0, 0, 0, 0, None))
    def compute_body_light_curve(surface_map, radius, x, y, z, time):
        theta = time.magnitude * 2 * jnp.pi / surface_map.period
        return (
            map_light_curve(
                surface_map,
                (system.central.radius / radius).magnitude,
                (x / radius).magnitude,
                (y / radius).magnitude,
                (z / radius).magnitude,
                theta,
            )
            * surface_map.amplitude
        )

    @quantity_input(time=ureg.day)
    @vectorize
    def light_curve_impl(time: Quantity) -> Array:
        xos, yos, zos = system.relative_position(time)

        if system.surface_map is None:
            central_light_curves = None
        else:
            theta = time.magnitude * 2 * jnp.pi / system.surface_map.period
            central_radius = system.central.radius
            central_light_curves = (
                central_bodies_lc(
                    system.surface_map,
                    (system.radius / central_radius).magnitude,
                    (xos / central_radius).magnitude,
                    (yos / central_radius).magnitude,
                    (zos / central_radius).magnitude,
                    theta,
                )
                * system.surface_map.amplitude
            )

        if all(surface_map is None for surface_map in system.surface_maps):
            body_light_curves = None
        else:
            body_light_curves = compute_body_light_curve(  # type: ignore
                system.radius, -xos, -yos, -zos, time
            )

        result = jnp.zeros(system.shape, dtype=time.dtype)
        if central_light_curves is not None:
            result += central_light_curves
        if body_light_curves is not None:
            result += body_light_curves

        return result

    return light_curve_impl


# TODO: figure out the sparse matrices (and Pijk) to avoid todense()
def map_light_curve(
    map: Optional[Map], r: Array, xo: Array, yo: Array, zo: Array, theta: Array
) -> Array:
    """Light curve of a map

    Args:
        map (Map): map object
        r (float): radius of the occulting body, relative to the current map body
        xo (float): x position of the occulting body, relative to the current map body
        yo (float): y position of the occulting body, relative to the current map body
        zo (float): z position of the occulting body, relative to the current map body
        theta (float): rotation angle of the map

    Returns:
        ArrayLike: light curve
    """
    if map is None:
        return jnp.zeros_like(theta)

    # TODO(lgrcia): Is this the right behavior when map.u is None?
    U = jnp.array([1, *(() if map.u is None else map.u)])
    b = jnp.sqrt(jnp.square(xo) + jnp.square(yo))
    b_rot = jnp.logical_or(jnp.greater_equal(b, 1.0 + r), jnp.less_equal(zo, 0.0))
    b_occ = jnp.logical_not(b_rot)

    # Occultation
    theta_z = jnp.arctan2(xo, yo)
    sT = solution_vector(map.deg)(b, r)
    # the reason for this if is that scipy.sparse.linalg.inv of a sparse matrix[[1]]
    # is a non-sparse [[1]], hence from_scipy_sparse raises an error (case deg=0) ...
    if map.deg > 0:
        A2 = scipy.sparse.linalg.inv(A2_inv(map.deg))
        A2 = sparse.BCOO.from_scipy_sparse(A2)
    else:
        A2 = jnp.array([1])

    sTA2 = sT @ A2

    # TODO(lgrcia): Is this the right behavior when map.y is None?
    if map.y is None:
        rotated_y = jnp.zeros(map.ydeg)
    else:
        rotated_y = left_project(
            map.ydeg, map.inc, map.obl, theta, theta_z, map.y.todense()
        )

    # limb darkening product
    A1_val = sparse.BCOO.from_scipy_sparse(A1(map.ydeg))
    p_y = Pijk.from_dense(A1_val @ rotated_y, degree=map.ydeg)
    p_u = Pijk.from_dense(U @ U0(map.udeg), degree=map.udeg)
    p_y = p_y * p_u

    x = jnp.where(b_occ, sTA2, rT(map.deg))
    norm = np.pi / (p_u.tosparse() @ rT(map.udeg))

    return (p_y.tosparse() @ x) * norm


def rT(lmax: int) -> Array:
    rt = [0.0 for _ in range((lmax + 1) * (lmax + 1))]
    amp0 = jnp.pi
    lfac1 = 1.0
    lfac2 = 2.0 / 3.0
    for ell in range(0, lmax + 1, 4):
        amp = amp0
        for m in range(0, ell + 1, 4):
            mu = ell - m
            nu = ell + m
            rt[ell * ell + ell + m] = amp * lfac1
            rt[ell * ell + ell - m] = amp * lfac1
            if ell < lmax:
                rt[(ell + 1) * (ell + 1) + ell + m + 1] = amp * lfac2
                rt[(ell + 1) * (ell + 1) + ell - m + 1] = amp * lfac2
            amp *= (nu + 2.0) / (mu - 2.0)
        lfac1 /= (ell / 2 + 2) * (ell / 2 + 3)
        lfac2 /= (ell / 2 + 2.5) * (ell / 2 + 3.5)
        amp0 *= 0.0625 * (ell + 2) * (ell + 2)

    amp0 = 0.5 * jnp.pi
    lfac1 = 0.5
    lfac2 = 4.0 / 15.0
    for ell in range(2, lmax + 1, 4):
        amp = amp0
        for m in range(2, ell + 1, 4):
            mu = ell - m
            nu = ell + m
            rt[ell * ell + ell + m] = amp * lfac1
            rt[ell * ell + ell - m] = amp * lfac1
            if ell < lmax:
                rt[(ell + 1) * (ell + 1) + ell + m + 1] = amp * lfac2
                rt[(ell + 1) * (ell + 1) + ell - m + 1] = amp * lfac2
            amp *= (nu + 2.0) / (mu - 2.0)
        lfac1 /= (ell / 2 + 2) * (ell / 2 + 3)
        lfac2 /= (ell / 2 + 2.5) * (ell / 2 + 3.5)
        amp0 *= 0.0625 * ell * (ell + 4)
    return np.array(rt)


def rTA1(lmax: int) -> Array:
    return rT(lmax) @ A1(lmax)



import math
from collections import defaultdict
from collections.abc import Mapping
from typing import Any, Optional

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from jax.experimental.sparse import BCOO
from scipy.special import legendre as LegendreP

from jaxoplanet.experimental.starry.rotation import dot_rotation_matrix
from jaxoplanet.experimental.starry.wigner3j import Wigner3jCalculator
from jaxoplanet.types import Array


class Ylm(eqx.Module):
    """Ylm object containing the spherical harmonic coefficients.

    Args:
        data (Mapping[tuple[int, int], Array], optional): dictionary of
            spherical harmonic coefficients. Defaults to {(0, 0): 1.0}.
        relative (bool, optional): Whether to normalize the coefficients by the
            value of the (0, 0) coefficient. Defaults to True.

    Raises:
        ValueError: if relative is True and the (0, 0) coefficient is zero.
    """

    # coefficients of the spherical harmonic expansion of the map in the form
    # {(l, m): coefficient}
    data: dict[tuple[int, int], Array]

    # maximum degree of the spherical harmonic expansion
    ell_max: int = eqx.field(static=True)

    # whether the spherical harmonic expansion is diagonal (all m=0)
    diagonal: bool = eqx.field(static=True)

    def __init__(
        self,
        data: Optional[Mapping[tuple[int, int], Array]] = None,
        normalize: bool = True,
    ):
        if data is None:
            data = {(0, 0): 1.0}

        if normalize:
            assert data[(0, 0)] != 0.0, ValueError(
                "The (0, 0) coefficient must be non-zero if relative=True"
            )
            data = {k: v / data[(0, 0)] for k, v in data.items()}
        self.data = dict(data)
        self.ell_max = max(ell for ell, _ in data.keys())
        self.diagonal = all(m == 0 for _, m in data.keys())

    @property
    def shape(self) -> tuple[int, ...]:
        """The number of coefficients in the expansion. This sets the shape of
        the output of `todense`."""
        return (self.ell_max**2 + 2 * self.ell_max + 1,)

    @property
    def indices(self) -> list[tuple[int, int]]:
        return list(self.data.keys())

    def index(self, ell: Array, m: Array) -> Array:
        """Convert the degree and order of the spherical harmonic to the
        corresponding index in the coefficient array."""
        return ell * (ell + 1) + m

    def tosparse(self) -> BCOO:
        indices, values = zip(*self.data.items())
        idx = jnp.array([self.index(ell, m) for ell, m in indices])[:, None]
        return BCOO((jnp.asarray(values), idx), shape=self.shape)

    def todense(self) -> Array:
        return self.tosparse().todense()

    @classmethod
    def from_dense(cls, y: Array, normalize: bool = True) -> "Ylm":
        data = {}
        for i, ylm in enumerate(y):
            ell = int(np.floor(np.sqrt(i)))
            m = i - ell * (ell + 1)
            data[(ell, m)] = ylm
        return cls(data, normalize=normalize)

    def __mul__(self, other: Any) -> "Ylm":
        if isinstance(other, Ylm):
            return _mul(self, other)
        else:
            return jax.tree_util.tree_map(lambda x: x * other, self)

    def __rmul__(self, other: Any) -> "Ylm":
        assert not isinstance(other, Ylm)
        return jax.tree_util.tree_map(lambda x: other * x, self)

    def __getitem__(self, key) -> Array:
        assert isinstance(key, tuple)
        return self.todense()[self.index(*key)]


def _mul(f: Ylm, g: Ylm) -> Ylm:
    """
    Based closely on the implementation from the MIT-licensed spherical package:

    https://github.com/moble/spherical/blob/0aa81c309cac70b90f8dfb743ce35d2cc9ae6dee/spherical/multiplication.py
    """
    ellmax_f = f.ell_max
    ellmax_g = g.ell_max
    ellmax_fg = ellmax_f + ellmax_g
    fg = defaultdict(lambda *_: 0.0)
    m_calculator = Wigner3jCalculator(ellmax_f, ellmax_g)
    for ell1 in range(ellmax_f + 1):
        sqrt1 = math.sqrt((2 * ell1 + 1) / (4 * math.pi))
        for m1 in range(-ell1, ell1 + 1):
            idx1 = (ell1, m1)
            if idx1 not in f.data:
                continue
            sum1 = sqrt1 * f.data[idx1]
            for ell2 in range(ellmax_g + 1):
                sqrt2 = math.sqrt(2 * ell2 + 1)
                # w3j_s = s_calculator.calculate(ell1, ell2, s_f, s_g)
                for m2 in range(-ell2, ell2 + 1):
                    idx2 = (ell2, m2)
                    if idx2 not in g.data:
                        continue
                    w3j_m = m_calculator.calculate(ell1, ell2, m1, m2)
                    sum2 = sqrt2 * g.data[idx2]
                    m3 = m1 + m2
                    for ell3 in range(
                        max(abs(m3), abs(ell1 - ell2)), min(ell1 + ell2, ellmax_fg) + 1
                    ):
                        # Could loop over same (ell3, m3) more than once, so add all
                        # contributions together
                        fg[(ell3, m3)] += (
                            (
                                math.pow(-1, ell1 + ell2 + ell3 + m3)
                                * math.sqrt(2 * ell3 + 1)
                                * w3j_m[ell3]  # Wigner3j(ell1, ell2, ell3, m1, m2, -m3)
                            )
                            * sum1
                            * sum2
                        )
    return Ylm(fg)


def Bp(ydeg, npts: int = 1000, eps: float = 1e-9, smoothing=None):
    """
    Return the matrix B+. This expands the
    spot profile `b` in Legendre polynomials. From https://github.com/rodluger/
    mapping_stellar_surfaces/blob/paper2-arxiv/paper2/figures/spot_profile.py and
    _spot_setup in starry/_core/core.py.
    """
    if smoothing is None:
        if ydeg < 4:
            smoothing = 0.5
        else:
            smoothing = 2.0 / ydeg

    theta = jnp.linspace(0, jnp.pi, npts)
    cost = jnp.cos(theta)
    B = jnp.hstack(
        [
            jnp.sqrt(2 * l + 1) * LegendreP(l)(cost).reshape(-1, 1)
            for l in range(ydeg + 1)
        ]
    )
    _Bp = jnp.linalg.solve(B.T @ B + eps * jnp.eye(ydeg + 1), B.T)
    l = jnp.arange(ydeg + 1)
    indices = l * (l + 1)
    S = jnp.exp(-0.5 * indices * smoothing**2)
    return (S[:, None] * _Bp, theta, indices)


def spot_profile(theta, radius, spot_fac=300):
    """
    The sigmoid spot profile.

    """
    z = spot_fac * (theta - radius)
    return 1 / (1 + jnp.exp(-z)) - 1


def ylm_spot(ydeg: int) -> callable:
    """spot expansion in the spherical harmonics basis.

    Args:
        ydeg (int): max degree of the spherical harmonics

    Returns:
        callable: function that returns the spherical harmonics coefficients of the spot
    """
    B, theta, indices = Bp(ydeg)

    def func(contrast: float, r: float, lat: float = 0.0, lon: float = 0.0):
        """spot expansion in the spherical harmonics basis.

        Args:
            contrast (float): spot contrast, defined as (1-c) where c is the intensity of
            the center of the spot relative to the unspotted surface. A contrast of 1.
            means that the spot intensity drops to zero at the center, 0. means that
            the intensity at the center of the spot is the same as the intensity of the
            unspotted surface.
            r (float): radius of the spot.
            lat (float, optional): latitude of the spot, assuming that the center of a
            star with an inclination of pi/2 has a latitude of 0. Defaults to 0.0.
            lon (float, optional): longitude of the spot, assuming that the center of a
            star with an inclination of pi/2 has a longitude of 0. Defaults to 0.0.

        Returns:
            Ylm: Ylm object containing the spherical harmonics coefficients of the spot
        """
        b = spot_profile(theta, r)
        y = jnp.zeros((ydeg + 1) * (ydeg + 1))
        y = y.at[indices].set(B @ b * contrast)
        y = y.at[0].set(y[0] + 1.0)
        y = dot_rotation_matrix(ydeg, 1.0, 0.0, 0.0, lat)(y)
        y = dot_rotation_matrix(ydeg, 0.0, 1.0, 0.0, -lon)(y)

        return Ylm.from_dense(y, normalize=False)

    return func


# TODO
def ring_y(l_max, pts=1000, eps=1e-9, smoothing=None):
    Bp, theta, idxs = None, None, None
    n_max = l_max**2 + 2 * l_max + 1

    def _y(contrast: float, width: float, latitude: float):
        b = 1 - jnp.array((theta > latitude - width) & (theta < latitude + width))
        y = jnp.zeros(n_max)
        y = y.at[idxs].set(Bp @ b)
        return y * contrast

    return _y



import numpy as np

from jaxoplanet.experimental.starry.maps import Map
from jaxoplanet.experimental.starry.utils import graticule
from jaxoplanet.experimental.starry.ylm import Ylm


def show_map(
    ylm_map_or_body,
    theta: float = 0.0,
    res: int = 400,
    n: int = 6,
    ax=None,
    white_contour: bool = True,
    **kwargs,
):
    """Show map of a

    Args:
        map_or_body (Map, Central or Body): Map or Body with a map
        theta (float, optional): Rotation angle of the map wrt its rotation axis.
        Defaults to 0.0.
        res (int, optional): Resolution of the map render. Defaults to 400.
        n (int, optional): number of latitude and longitude lines to show.
        Defaults to 6.
        ax (matplotlib.pyplot.Axes, optional): plot axes. Defaults to None.
        white_contour (bool, optional): Whether to surround the map by a white border
        (to hide border pixel aliasing). Defaults to True.
    """
    import matplotlib.pyplot as plt

    if ax is None:
        ax = plt.gca()
        if ax is None:
            ax = plt.subplot(111)

    if hasattr(ylm_map_or_body, "map"):
        map = ylm_map_or_body.map
        radius = ylm_map_or_body.radius.magnitude
        n = int(np.ceil(n * np.cbrt(radius)))
    # import Ylm leads to circular import
    elif isinstance(ylm_map_or_body, Ylm):
        map = Map(y=ylm_map_or_body)
        radius = 1.0
    else:
        map = ylm_map_or_body
        radius = 1.0

    ax.imshow(
        map.render(
            theta,
            res,
        ),
        origin="lower",
        **kwargs,
        extent=(-radius, radius, -radius, radius),
    )
    if n is not None:
        graticule(
            map.inc,
            map.obl,
            theta,
            radius=radius,
            n=n,
            white_contour=white_contour,
        )
    ax.axis(False)



from functools import partial
from typing import Optional

import equinox as eqx
import jax
import jax.numpy as jnp
from jax.scipy.spatial.transform import Rotation

from jaxoplanet.experimental.starry.basis import A1, U0, poly_basis
from jaxoplanet.experimental.starry.pijk import Pijk
from jaxoplanet.experimental.starry.rotation import (
    left_project,
    right_project_axis_angle,
)
from jaxoplanet.experimental.starry.utils import ortho_grid
from jaxoplanet.experimental.starry.ylm import Ylm


class Map(eqx.Module):
    """Surface map object.

    Args:
        y (Optional[Union[Ylm, float]], optional): Ylm object containing the
        spherical harmonic expansion of the map. Defaults to None (a uniform map)
        with amplitude 1.0.
        inc (Optional[float], optional): inclination of the map. Defaults to 0.
        obl (Optional[float], optional): obliquity iof the map. Defaults to 0.
        u (Optional[tuple], optional): polynomial limb-darkening coefficients of
        the map. Defaults to ().
        period (Optional[float], optional): rotation period of the map. Defaults to
        1e15.
        amplitude (Optional[float], optional): amplitude of the map, this quantity is
        proportional to the luminosity of the map and multiplies all flux-related
        observables. Defaults to 1..
        normalize (Optional[bool], optional): whether to normalize the coefficients
        of the spherical harmonics. If None or True, Ylm is normalized and the
        amplitude of the map is set to y[(0, 0)]. Defaults to None.

    Example:

        .. code-block:: python

            import numpy as np
            import jax
            from jaxoplanet.experimental.starry.utils import show_map
            from jaxoplanet.experimental.starry.maps import Map
            from jaxoplanet.experimental.starry.ylm import Ylm

            jax.config.update("jax_enable_x64", True)

            np.random.seed(30)
            y = Ylm.from_dense(np.random.rand(20))
            m = Map(y=y, u=[0.5, 0.1], inc=0.9, obl=-0.3)
            show_map(m)
    """

    # Ylm object representing the spherical harmonic expansion of the map.
    y: Optional[Ylm] = None

    # Inclination of the map in radians.
    inc: Optional[float] = None

    # Obliquity of the map in radians.
    obl: Optional[float] = None

    # Tuple of limb darkening coefficients.
    u: Optional[tuple] = None

    # Rotation period of the map in days (attribute subject to change)
    period: Optional[float] = None

    # Amplitude of the map, a quantity proportional to map luminosity.
    amplitude: Optional[float] = None

    # Whether to normalize the coefficients of the spherical harmonic.
    normalize: Optional[bool] = None

    def __init__(
        self,
        *,
        y: Optional[Ylm] = None,
        inc: Optional[float] = None,
        obl: Optional[float] = None,
        u: Optional[tuple] = None,
        period: Optional[float] = None,
        amplitude: Optional[float] = None,
        normalize: Optional[bool] = None,
    ):
        if y is None:
            y = Ylm()
        if inc is None:
            inc = jnp.pi / 2
        if obl is None:
            obl = 0.0
        if u is None:
            u = ()
        if period is None:
            period = 1e15
        if amplitude is None:
            amplitude = 1.0
        if normalize is None:
            normalize = True

        if normalize:
            amplitude = y[(0, 0)]
            y = Ylm(data=y.data, normalize=True)

        self.y = y
        self.inc = inc
        self.obl = obl
        self.u = u
        self.period = period
        self.amplitude = amplitude

    @property
    def poly_basis(self):
        return jax.jit(poly_basis(self.deg))

    @property
    def udeg(self):
        return len(self.u)

    @property
    def ydeg(self):
        return self.y.ell_max

    @property
    def deg(self):
        return self.ydeg + self.udeg

    def _intensity(self, x, y, z, theta=0.0):
        pT = self.poly_basis(x, y, z)
        Ry = left_project(self.ydeg, self.inc, self.obl, theta, 0.0, self.y.todense())
        A1Ry = A1(self.ydeg).todense() @ Ry
        p_y = Pijk.from_dense(A1Ry, degree=self.ydeg)
        U = jnp.array([1, *self.u])
        p_u = Pijk.from_dense(U @ U0(self.udeg), degree=self.udeg)
        p = (p_y * p_u).todense()
        return pT @ p * self.amplitude

    @partial(jax.jit, static_argnames=("res",))
    def render(self, theta: float = 0.0, res: int = 400):
        """Returns the intensity map projected onto the x-y plane (sky).

        Args:
            theta (float, optional): rotation angle of the map. Defaults to 0.0.
            res (int, optional): resolution of the render. Defaults to 400.

        Returns:
            ArrayLike: square 2D array representing the intensity map
            (with nans outside the map disk).
        """
        _, xyz = ortho_grid(res)
        intensity = self._intensity(*xyz, theta=theta)
        return jnp.reshape(intensity, (res, res))

    def intensity(self, lat: float, lon: float):
        """Returns the intensity of the map at a given latitude and longitude.

        Args:
            lat (float): latitude in the rest frame of the map
            lon (float): longitude in the rest frame of the map

        Returns:
            float: intensity of the map at the given latitude and longitude
        """
        lon = lon + jnp.pi / 2  # convention, 0 lon faces the observer
        lat = jnp.pi / 2 - lat  # convention, latitude 0 is equator
        x = jnp.sin(lat) * jnp.cos(lon)
        y = jnp.sin(lat) * jnp.sin(lon)
        z = jnp.cos(lat) * jnp.ones_like(x)

        axis = right_project_axis_angle(self.inc - jnp.pi / 2, self.obl, 0.0, 0.0)
        axis = jnp.array(axis[0:3]) * axis[-1]
        rotation = Rotation.from_rotvec(axis)
        x, y, z = rotation.apply(jnp.array([x, y, z]).T).T

        return self._intensity(x, y, z)

    def flux(self, time):
        from jaxoplanet.experimental.starry.light_curves import map_light_curve

        theta = time * 2 * jnp.pi / self.period
        return (
            jnp.vectorize(partial(map_light_curve, self, 0.0, 2.0, 2.0, 2.0))(theta)
            * self.amplitude
        )



from functools import partial
from typing import Callable

import jax
import jax.numpy as jnp
import numpy as np
from scipy.special import roots_legendre

from jaxoplanet.core.limb_dark import kite_area
from jaxoplanet.types import Array


def solution_vector(l_max: int, order: int = 20) -> Callable[[Array, Array], Array]:
    n_max = l_max**2 + 2 * l_max + 1

    @jax.jit
    @partial(jnp.vectorize, signature=f"(),()->({n_max})")
    def impl(b: Array, r: Array) -> Array:
        b = jnp.abs(b)
        r = jnp.abs(r)
        kappa0, kappa1 = kappas(b, r)
        P = p_integral(order, l_max, b, r, kappa0)
        Q = q_integral(l_max, 0.5 * jnp.pi - kappa1)
        return Q - P

    return impl


def kappas(b: Array, r: Array) -> tuple[Array, Array]:
    b2 = jnp.square(b)
    factor = (r - 1) * (r + 1)
    b_cond = jnp.logical_and(jnp.greater(b, jnp.abs(1 - r)), jnp.less(b, 1 + r))
    b_ = jnp.where(b_cond, b, 1)
    area = jnp.where(b_cond, kite_area(r, b_, 1), 0)
    return jnp.arctan2(area, b2 + factor), jnp.arctan2(area, b2 - factor)


def q_integral(l_max: int, lam: Array) -> Array:
    zero = jnp.zeros_like(lam)
    c = jnp.cos(lam)
    s = jnp.sin(lam)
    h = {
        (0, 0): 2 * lam + jnp.pi,
        (0, 1): -2 * c,
    }

    def get(u: int, v: int) -> Array:
        if (u, v) in h:
            return h[(u, v)]
        if u >= 2:
            comp = 2 * c ** (u - 1) * s ** (v + 1) + (u - 1) * get(u - 2, v)
        else:
            assert v >= 2
            comp = -2 * c ** (u + 1) * s ** (v - 1) + (v - 1) * get(u, v - 2)
        comp /= u + v
        h[(u, v)] = comp
        return comp

    U = []
    for l in range(l_max + 1):  # noqa
        for m in range(-l, l + 1):
            if l == 1 and m == 0:
                U.append((np.pi + 2 * lam) / 3)
                continue
            mu = l - m
            nu = l + m
            if (mu % 2) == 0 and (mu // 2) % 2 == 0:
                u = mu // 2 + 2
                v = nu // 2
                assert u % 2 == 0
                U.append(get(u, v))
            else:
                U.append(zero)

    return jnp.stack(U)


def p_integral(order: int, l_max: int, b: Array, r: Array, kappa0: Array) -> Array:
    b2 = jnp.square(b)
    r2 = jnp.square(r)

    # This is a hack for when r -> 0 or b -> 0, so k2 -> inf
    factor = 4 * b * r
    k2_cond = jnp.less(factor, 10 * jnp.finfo(factor.dtype).eps)
    factor = jnp.where(k2_cond, 1, factor)
    k2 = jnp.maximum(0, (1 - r2 - b2 + 2 * b * r) / factor)

    # And for when r -> 0
    r_cond = jnp.less(r, 10 * jnp.finfo(r.dtype).eps)
    delta = (b - r) / (2 * jnp.where(r_cond, 1, r))

    roots, weights = roots_legendre(order)
    rng = 0.5 * kappa0
    phi = rng * roots
    c = jnp.cos(phi + 0.5 * kappa0)
    s = jnp.sin(phi)
    s2 = jnp.square(s)

    f0 = jnp.maximum(0, jnp.where(k2_cond, 1 - r2, factor * (k2 - s2))) ** 1.5
    a1 = s2 - jnp.square(s2)
    a2 = jnp.where(r_cond, 0, delta + s2)
    a4 = 1 - 2 * s2

    ind = []
    arg = []
    n = 0
    for l in range(l_max + 1):  # noqa
        fa3 = (2 * r) ** (l - 1) * f0
        for m in range(-l, l + 1):
            mu = l - m
            nu = l + m

            if mu == 1 and l == 1:
                omz2 = r2 + b2 - 2 * b * r * c
                cond = jnp.less(omz2, 10 * jnp.finfo(omz2.dtype).eps)
                omz2 = jnp.where(cond, 1, omz2)
                z2 = jnp.maximum(0, 1 - omz2)
                result = 2 * r * (r - b * c) * (1 - z2 * jnp.sqrt(z2)) / (3 * omz2)
                arg.append(jnp.where(cond, 0, result))

            elif mu % 2 == 0 and (mu // 2) % 2 == 0:
                arg.append(
                    2 * (2 * r) ** (l + 2) * a1 ** (0.25 * (mu + 4)) * a2 ** (0.5 * nu)
                )

            elif mu == 1 and l % 2 == 0:
                arg.append(fa3 * a1 ** (l // 2 - 1) * a4)

            elif mu == 1:
                arg.append(fa3 * a1 ** ((l - 3) // 2) * a2 * a4)

            elif (mu - 1) % 2 == 0 and ((mu - 1) // 2) % 2 == 0:
                arg.append(2 * fa3 * a1 ** ((mu - 1) // 4) * a2 ** (0.5 * (nu - 1)))

            else:
                n += 1
                continue

            ind.append(n)
            n += 1

    P0 = rng * jnp.sum(jnp.stack(arg) * weights[None, :], axis=1)
    P = jnp.zeros(l_max**2 + 2 * l_max + 1)

    # Yes, using np not jnp here: 'ind' is always static.
    inds = np.stack(ind)

    return P.at[inds].set(P0)



from collections.abc import Iterable, Sequence
from typing import Any, Callable, Optional, Union

from jaxoplanet.experimental.starry.maps import Map
from jaxoplanet.object_stack import ObjectStack
from jaxoplanet.orbits.keplerian import Body, Central, OrbitalBody, System


class SurfaceMapBody(Body):
    surface_map: Optional[Map] = None


class SurfaceMapSystem(System):
    surface_map: Optional[Map]
    _surface_map_stack: ObjectStack[Map]

    def __init__(
        self,
        central: Optional[Central] = None,
        surface_map: Optional[Map] = None,
        *,
        bodies: Iterable[
            tuple[Union[Body, OrbitalBody, SurfaceMapBody], Optional[Map]]
        ] = (),
    ):
        self.central = Central() if central is None else central
        self.surface_map = surface_map

        orbital_bodies = []
        surface_maps = []
        for body, surface_map in bodies:
            if isinstance(body, OrbitalBody):
                orbital_bodies.append(body)
                surface_maps.append(surface_map)
            else:
                orbital_bodies.append(OrbitalBody(self.central, body))
                if surface_map is None:
                    surface_maps.append(getattr(body, "surface_map", None))
                else:
                    surface_maps.append(surface_map)

        self._body_stack = ObjectStack(*orbital_bodies)
        self._surface_map_stack = ObjectStack(*surface_maps)

    @property
    def surface_maps(self) -> tuple[Map, ...]:
        return self._surface_map_stack.objects

    def add_body(
        self,
        body: Optional[Union[Body, SurfaceMapBody]] = None,
        surface_map: Optional[Map] = None,
        **kwargs: Any,
    ) -> "SurfaceMapSystem":
        if body is None:
            body = Body(**kwargs)
        if surface_map is None:
            surface_map = getattr(body, "surface_map", None)
        bodies = list(zip(self.bodies, self.surface_maps)) + [(body, surface_map)]
        return SurfaceMapSystem(
            central=self.central, surface_map=self.surface_map, bodies=bodies
        )

    def surface_map_vmap(
        self,
        func: Callable,
        in_axes: Union[int, None, Sequence[Any]] = 0,
        out_axes: Any = 0,
    ) -> Callable:
        return self._surface_map_stack.vmap(func, in_axes=in_axes, out_axes=out_axes)



from jaxoplanet.experimental.starry.maps import Map as Map
from jaxoplanet.experimental.starry.solution import (
    solution_vector as solution_vector,
)
from jaxoplanet.experimental.starry.visualization import show_map as show_map
from jaxoplanet.experimental.starry.ylm import Ylm as Ylm
